from .create_command import get_create_command
from .deploy_command import get_deploy_command
from .get_command import get_get_command
from .init_command import get_init_command
from .list_command import get_list_command
from .login_command import get_login_command
from .logout_command import get_logout_command
from .remove_command import get_remove_command
from .use_command import get_set_command



import sys


class OutputCallBack:
    def print_header(self, line):
        sys.stdout.write(f"{line}\n")

    def print_line(self, line):
        sys.stdout.write(f"{line}\n")

    def print_lines_in_panel(self, lines, header):
        sys.stdout.write(header)
        for line in lines:
            sys.stdout.write(line)



import logging

import rich_click as click

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib import deployment as deployment_lib
from servicefoundry.build.lib import service as service_lib
from servicefoundry.build.lib import workspace as workspace_lib
from servicefoundry.build.lib.messages import (
    PROMPT_NO_DEPLOYMENTS,
    PROMPT_NO_SERVICES,
    PROMPT_NO_WORKSPACES,
)
from servicefoundry.build.model.entity import (
    Cluster,
    Deployment,
    Secret,
    SecretGroup,
    Service,
    Workspace,
)
from servicefoundry.cli.const import (
    ENABLE_AUTHORIZE_COMMANDS,
    ENABLE_CLUSTER_COMMANDS,
    ENABLE_SECRETS_COMMANDS,
)
from servicefoundry.cli.display_util import print_list
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.group(name="list")
def list_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Servicefoundry list resources

    \b
    Supported resources:
    - workspace
    - service
    - deployment
    """
    pass


@click.command(name="cluster", help="list cluster")
@handle_exception_wrapper
def list_cluster():
    tfs_client = ServiceFoundryServiceClient.get_client()
    clusters = tfs_client.list_cluster()
    print_list("Clusters", clusters, Cluster.display_columns)


@click.command(name="workspace", help="list workspaces")
@click.option("-A", "--all", is_flag=True, default=False)
@click.option("-c", "--cluster", type=click.STRING, default=None, help="cluster name")
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def list_workspace(all, cluster, non_interactive):
    # Tests:
    # - Set Context -> list workspace -> Should get workspaces in set cluster
    # - Set Context -> list workspace -c 'cluster_name' -> Should get workspaces in given cluster
    # - Set Context -> list workspace -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - Set Context -> list workspace -A -> Should give all workspaces across all clusters
    # - No Context -> list workspace -c 'cluster_name' -> Should get workspaces in given cluster
    # - No Context -> list workspace -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - No Context -> list workspace -A -> Should give all workspaces across all clusters
    # ? No Context -> list workspace -> Should list workspaces if there is only cluster or ask for cluster name
    workspaces = workspace_lib.list_workspaces(
        cluster_name_or_id=cluster,
        all_=all,
        non_interactive=non_interactive,
    )
    if not workspaces:
        console.print(PROMPT_NO_WORKSPACES)
    else:
        workspaces.sort(key=lambda w: (w.fqn, w.createdAt))
    # TODO (chiragjn): Display columns here need to show cluster name!
    print_list(
        "Workspaces",
        [w.to_dict() for w in workspaces],
        columns=Workspace.display_columns,
    )


@click.command(name="service", help="list service in a workspace")
@click.option("-A", "--all", is_flag=True, default=False)
@click.option(
    "-w", "--workspace", type=click.STRING, default=None, help="workspace name"
)
@click.option("-c", "--cluster", type=click.STRING, default=None, help="cluster name")
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def list_service(all, workspace, cluster, non_interactive):
    services = service_lib.list_services(
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        all_=all,
        non_interactive=non_interactive,
    )
    if not services:
        console.print(PROMPT_NO_SERVICES)
    else:
        services.sort(key=lambda s: (s.fqn, s.name))
    # TODO (chiragjn): Display columns here need to show workspace and cluster name!
    print_list(
        f"Services",
        [s.to_dict() for s in services],
        columns=Service.display_columns,
    )


@click.command(name="deployment", help="list deployment")
@click.argument("service")
@click.option(
    "-w", "--workspace", type=click.STRING, default=None, help="workspace name"
)
@click.option("-c", "--cluster", type=click.STRING, default=None, help="cluster name")
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def list_deployment(service, workspace, cluster, non_interactive):
    deployments = deployment_lib.list_deployments(
        service_name_or_id=service,
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        all_=False,  # TODO (chirajn): Add support for all flag
        non_interactive=non_interactive,
    )
    if not deployments:
        console.print(PROMPT_NO_DEPLOYMENTS)
    else:
        deployments.sort(key=lambda d: (d.fqn, d.name))
    print_list(
        f"Deployments",
        [d.to_dict() for d in deployments],
        columns=Deployment.display_columns,
    )


@click.command(name="secret-group", help="list secret groups")
@handle_exception_wrapper
def list_secret_group():
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.get_secret_groups()
    print_list("Secret Groups", response, columns=SecretGroup.display_columns)


@click.command(name="secret", help="list secrets in a group")
@click.argument("secret_group_id")
@handle_exception_wrapper
def list_secret(secret_group_id):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.get_secrets_in_group(secret_group_id)
    print_list("Secrets", response, columns=Secret.display_columns)


@click.command(name="authorize", help="list authorization for a resource id.")
@click.argument("resource_type", type=click.Choice(["workspace"], case_sensitive=False))
@click.argument("resource_id")
@handle_exception_wrapper
def list_authorize(resource_type, resource_id):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.get_authorization_for_resource(resource_type, resource_id)
    print_list(f"Auth for {resource_type}: {resource_id}", response)


def get_list_command():
    list_command.add_command(list_workspace)
    list_command.add_command(list_service)
    list_command.add_command(list_deployment)

    if ENABLE_AUTHORIZE_COMMANDS:
        list_command.add_command(list_authorize)
    if ENABLE_CLUSTER_COMMANDS:
        list_command.add_command(list_cluster)
    if ENABLE_SECRETS_COMMANDS:
        list_command.add_command(list_secret)
        list_command.add_command(list_secret_group)

    return list_command



import logging

import rich_click as click

from servicefoundry.build.lib import context as context_lib
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.group(name="use")
def use_command():
    """
    Set default workspace

    \b
    Supported resources:
    - workspace
    - cluster
    """
    pass


@click.command(name="cluster", help="Set default cluster")
@click.argument("cluster", type=click.STRING, required=False, default=None)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def use_cluster(cluster, non_interactive):
    context_lib.use_cluster(name_or_id=cluster, non_interactive=non_interactive)


@click.command(name="workspace", help="Set default workspace")
@click.argument("workspace", type=click.STRING, required=False, default=None)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def use_workspace(workspace, non_interactive):
    context_lib.use_workspace(
        name_or_id=workspace,
        cluster_name_or_id=None,  # pick from context if available
        non_interactive=non_interactive,
    )


def get_set_command():
    use_command.add_command(use_cluster)
    use_command.add_command(use_workspace)
    return use_command



from typing import Optional

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib.messages import (
    PROMPT_SETTING_CLUSTER_CONTEXT,
    PROMPT_SETTING_WORKSPACE_CONTEXT,
)
from servicefoundry.build.lib.util import (
    resolve_cluster_or_error,
    resolve_workspace_or_error,
)
from servicefoundry.build.model.entity import Cluster, Workspace


def use_cluster(
    name_or_id: Optional[str] = None,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Cluster:
    client = client or ServiceFoundryServiceClient.get_client()
    cluster = resolve_cluster_or_error(
        name_or_id=name_or_id,
        non_interactive=non_interactive,
        ignore_context=True,
        client=client,
    )
    client.session.set_cluster(cluster.to_dict_for_session())
    console.print(PROMPT_SETTING_CLUSTER_CONTEXT.format(cluster.name))
    client.session.save_session()
    return cluster


def use_workspace(
    name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[str] = None,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient.get_client()
    if non_interactive:
        if not name_or_id:
            raise ValueError("workspace name or id cannot be null")

    cluster = resolve_cluster_or_error(
        name_or_id=cluster_name_or_id,
        ignore_context=False,
        non_interactive=non_interactive,
        client=client,
    )

    workspace, cluster = resolve_workspace_or_error(
        name_or_id=name_or_id,
        cluster_name_or_id=cluster,
        ignore_context=True,
        non_interactive=non_interactive,
        client=client,
    )
    client.session.set_workspace(workspace.to_dict_for_session())
    console.print(PROMPT_SETTING_WORKSPACE_CONTEXT.format(workspace.name))
    client.session.set_cluster(cluster.to_dict_for_session())
    console.print(PROMPT_SETTING_CLUSTER_CONTEXT.format(cluster.name))
    client.session.save_session()
    return workspace



import logging

import rich_click as click

from servicefoundry.build import lib
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.command()
@handle_exception_wrapper
def logout():
    """
    Logout servicefoundry session
    """
    lib.logout()


def get_logout_command():
    return logout



import functools
import zipfile

import rich_click as click
from questionary import Choice
from requests.exceptions import ConnectionError
from rich.padding import Padding
from rich.panel import Panel

from servicefoundry.build.console import console

from ..build.exceptions import BadRequestException, ConfigurationException
from .const import IS_DEBUG


def handle_exception(exception):
    if isinstance(exception, BadRequestException):
        print_error(
            f"[cyan bold]status_code[/] {exception.status_code}\n"
            f"[cyan bold]message[/]     {exception.message}"
        )
    elif isinstance(exception, ConnectionError):
        print_error(f"Couldn't connect to Servicefoundry.")
    elif isinstance(exception, ConfigurationException):
        print_error(f"[cyan bold]message[/]     {exception.message}")
    else:
        if IS_DEBUG:
            console.print_exception()
        else:
            console.print(f"[red][bold]Error:[/] {str(exception)}[/]")


def handle_exception_wrapper(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            handle_exception(e)

    return inner


def print_error(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="red",
            title="Command failed",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def print_message(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="cyan",
            title="Success",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def unzip_package(path_to_package, destination):
    with zipfile.ZipFile(path_to_package, "r") as zip_ref:
        zip_ref.extractall(destination)



import logging
from threading import Thread

import rich_click as click

from servicefoundry.build.build import LOCAL, REMOTE, build_and_deploy
from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.display_util import print_obj
from servicefoundry.cli.rich_output_callback import RichOutputCallBack
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.command(help="Deploy servicefoundry service")
@click.option("--local", is_flag=True, default=False)
@click.option("--env", "-e", type=click.STRING)
@click.argument("service_dir", type=click.Path(exists=True), nargs=1, default="./")
@handle_exception_wrapper
def deploy(local, env, service_dir):
    build = LOCAL if local else REMOTE
    # @TODO Give ability to user confirmation.
    # @TODO Pick servicefoundry.{env}.yaml based on env
    deployment = build_and_deploy(
        base_dir=service_dir,
        build=build,
        callback=RichOutputCallBack(),
    )
    if isinstance(deployment, Thread):
        deployment.join()
    else:
        print_obj("Deployment", deployment)
        if not CliConfig.get("json"):
            # tail logs
            tfs_client = ServiceFoundryServiceClient.get_client()
            tfs_client.tail_logs(deployment["runId"], wait=True)


def get_deploy_command():
    return deploy



import logging

import rich_click as click

from servicefoundry.build import lib
from servicefoundry.build.console import console
from servicefoundry.build.lib.messages import PROMPT_POST_LOGIN
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.command()
@handle_exception_wrapper
def login():
    """
    Login to servicefoundry

    \b
    Once logged in, you can initiate a new service with `sfy init`
    and deploy the service with `sfy deploy .`
    """
    # TODO (chiragjn): Add support for non interactive login with API key.
    #                  It is supported indirectly as we always look for `SERVICE_FOUNDRY_API_KEY`
    #                  in the environment
    lib.login(non_interactive=False)
    console.print(PROMPT_POST_LOGIN)


def get_login_command():
    return login



import time
from typing import Tuple

from prometheus_client import Counter, Gauge, Histogram
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Match
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR
from starlette.types import ASGIApp


class PrometheusMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, filter_unhandled_paths: bool = False) -> None:
        super().__init__(app)
        self.filter_unhandled_paths = filter_unhandled_paths
        self.REQUESTS = Counter(
            "request",
            "Counter of requests by method and path.",
            ["method", "path_template"],
        )
        self.RESPONSES = Counter(
            "response",
            "Counter of responses by method, path and status codes.",
            ["method", "path_template", "status_code"],
        )
        self.REQUESTS_PROCESSING_TIME = Histogram(
            "latency_in_sec",
            "Histogram of requests processing time by path (in seconds)",
            ["method", "path_template"],
        )
        self.EXCEPTIONS = Counter(
            "exception",
            "Count of exception raised by path and exception type",
            ["method", "path_template", "exception_type"],
        )
        self.REQUESTS_IN_PROGRESS = Gauge(
            "concurrent_request",
            "Concurrent request by method and path currently being processed",
            ["method", "path_template"],
        )

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        method = request.method
        path_template, is_handled_path = self.get_path_template(request)

        if self._is_path_filtered(is_handled_path):
            return await call_next(request)

        self.REQUESTS_IN_PROGRESS.labels(
            method=method, path_template=path_template
        ).inc()
        self.REQUESTS.labels(method=method, path_template=path_template).inc()
        before_time = time.perf_counter()
        try:
            response = await call_next(request)
        except BaseException as e:
            status_code = HTTP_500_INTERNAL_SERVER_ERROR
            self.EXCEPTIONS.labels(
                method=method,
                path_template=path_template,
                exception_type=type(e).__name__,
            ).inc()
            raise e from None
        else:
            status_code = response.status_code
            after_time = time.perf_counter()
            self.REQUESTS_PROCESSING_TIME.labels(
                method=method, path_template=path_template
            ).observe(after_time - before_time)
        finally:
            self.RESPONSES.labels(
                method=method, path_template=path_template, status_code=status_code
            ).inc()
            self.REQUESTS_IN_PROGRESS.labels(
                method=method, path_template=path_template
            ).dec()

        return response

    @staticmethod
    def get_path_template(request: Request) -> Tuple[str, bool]:
        for route in request.app.routes:
            match, child_scope = route.matches(request.scope)
            if match == Match.FULL:
                return route.path, True

        return request.url.path, False

    def _is_path_filtered(self, is_handled_path: bool) -> bool:
        return self.filter_unhandled_paths and not is_handled_path



import logging

import questionary
import rich_click as click
from rich import print_json

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.lib import service as service_lib
from servicefoundry.build.lib import workspace as workspace_lib
from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import (
    ENABLE_AUTHORIZE_COMMANDS,
    ENABLE_CLUSTER_COMMANDS,
    ENABLE_SECRETS_COMMANDS,
)
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.group(name="remove")
def remove_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Servicefoundry remove entity by id

    \b
    Supported resources:
    - workspace
    - service
    """
    pass


@click.command(name="cluster", help="remove cluster")
@click.argument("cluster_id")
@handle_exception_wrapper
def remove_cluster(cluster_id):
    tfs_client = ServiceFoundryServiceClient.get_client()
    confirm = questionary.confirm(
        default=False,
        auto_enter=False,
        message=f"Are you sure you want to remove cluster {cluster_id!r}?",
    ).ask()
    if confirm:
        tfs_client.remove_cluster(cluster_id)
        # Remove workspace and cluster from context
        ctx_cluster = tfs_client.session.get_cluster()
        if ctx_cluster and ctx_cluster["id"] == cluster_id:
            tfs_client.session.set_workspace(None)
            tfs_client.session.set_cluster(None)
            tfs_client.session.save_session()
    else:
        raise Exception("Aborted!")


@click.command(name="workspace", help="remove workspace")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to remove the workspace from",
)
@click.confirmation_option(prompt="Are you sure you want to remove this workspace?")
@handle_exception_wrapper
def remove_workspace(name, cluster):
    # Tests:
    # - Set Context -> remove workspace -> Should give error to give workspace name
    # - Set Context -> remove workspace valid_name -> Should delete
    # - Set Context -> remove workspace invalid_name -> Should give error no such workspace in set cluster
    # - Set Context -> remove workspace name -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - Set Context -> remove workspace invalid_name -c 'cluster_name' -> Should give error invalid workspace
    # - Set Context -> remove workspace valid_name -c 'cluster_name' -> Should delete
    # - No Context -> remove workspace -> Should give error to give workspace name
    # - No Context -> remove workspace valid_name -> Try to resolve, if only one exists then delete
    #                 otherwise error to give cluster
    # - No Context -> remove workspace invalid_name -> Tries to resolve, if only one exists then delete
    #                 otherwise error to give cluster
    # - No Context -> remove workspace name -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - No Context -> remove workspace invalid_name -c 'cluster_name' -> Should give error invalid workspace
    # - No Context -> remove workspace valid_name -c 'cluster_name' -> Should delete
    tail_logs = not CliConfig.get("json")
    response = workspace_lib.remove_workspace(
        name_or_id=name,
        cluster_name_or_id=cluster,
        tail_logs=tail_logs,
        non_interactive=True,
    )
    if not tail_logs:
        print_json(data=response)


@click.command(name="service", help="remove service")
@click.argument("name", type=click.STRING)
@click.option(
    "-w",
    "--workspace",
    type=click.STRING,
    default=None,
    help="workspace to find this service in",
)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to find this cluster in",
)
@click.confirmation_option(prompt="Are you sure you want to remove this service?")
@handle_exception_wrapper
def remove_service(name, workspace, cluster):
    tail_logs = not CliConfig.get("json")
    response = service_lib.remove_service(
        name_or_id=name,
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        tail_logs=tail_logs,
        non_interactive=True,
    )
    if not tail_logs:
        print_json(data=response)


@click.command(name="secret-group", help="remove secret-group")
@click.argument("secret_group_id")
@handle_exception_wrapper
def remove_secret_group(secret_group_id):
    confirm = questionary.confirm(
        default=False,
        auto_enter=False,
        message=f"Are you sure you want to remove secret group {secret_group_id!r}?",
    ).ask()
    if confirm:
        tfs_client = ServiceFoundryServiceClient.get_client()
        response = tfs_client.delete_secret_group(secret_group_id)
        print_json(data=response)
    else:
        raise Exception("Aborted!")


@click.command(name="secret", help="remove secret")
@click.argument("secret_id")
@handle_exception_wrapper
def remove_secret(secret_id):
    confirm = questionary.confirm(
        default=False,
        auto_enter=False,
        message=f"Are you sure you want to remove secret {secret_id!r}?",
    ).ask()
    if confirm:
        tfs_client = ServiceFoundryServiceClient.get_client()
        response = tfs_client.delete_secret(secret_id)
        print_json(data=response)
    else:
        raise Exception("Aborted!")


@click.command(name="auth", help="remove authorization")
@click.argument("authorization_id")
@handle_exception_wrapper
def remove_auth(authorization_id):
    confirm = questionary.confirm(
        default=False,
        auto_enter=False,
        message=f"Are you sure you want to remove authorization {authorization_id!r}?",
    ).ask()
    if confirm:
        tfs_client = ServiceFoundryServiceClient.get_client()
        response = tfs_client.delete_authorization(authorization_id)
        print_json(data=response)
    else:
        raise Exception("Aborted!")


def get_remove_command():
    remove_command.add_command(remove_workspace)
    remove_command.add_command(remove_service)

    if ENABLE_AUTHORIZE_COMMANDS:
        remove_command.add_command(remove_auth)

    if ENABLE_CLUSTER_COMMANDS:
        remove_command.add_command(remove_cluster)

    if ENABLE_SECRETS_COMMANDS:
        remove_command.add_command(remove_secret)
        remove_command.add_command(remove_secret_group)

    return remove_command



import logging

import rich_click as click
from rich import print_json

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib import workspace as workspace_lib
from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import ENABLE_CLUSTER_COMMANDS, ENABLE_SECRETS_COMMANDS
from servicefoundry.cli.display_util import print_obj
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)

WORKSPACE_DISPLAY_FIELDS = [
    "id",
    "name",
    "namespace",
    "status",
    "clusterId",
    "createdBy",
    "createdAt",
    "updatedAt",
]
DEPLOYMENT_DISPLAY_FIELDS = [
    "id",
    "serviceId",
    "domain",
    "deployedBy",
    "createdAt",
    "updatedAt",
]


@click.group(name="create")
def create_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Create servicefoundry resources

    \b
    Supported resources:
    - workspace
    """
    pass


@click.command(name="cluster", help="create new cluster")
@click.argument("name")
@click.argument("region")
@click.argument("aws_account_id")
@click.argument("server_name")
@click.argument("ca_data")
@click.argument("server_url")
@handle_exception_wrapper
def create_cluster(name, region, aws_account_id, server_name, ca_data, server_url):
    tfs_client = ServiceFoundryServiceClient.get_client()
    cluster = tfs_client.create_cluster(
        name, region, aws_account_id, server_name, ca_data, server_url
    )
    print_obj("Cluster", cluster)


@click.command(name="workspace", help="create new workspace in cluster")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to create this workspace in",
)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def create_workspace(name, cluster, non_interactive):
    tail_logs = not CliConfig.get("json")
    workspace = workspace_lib.create_workspace(
        name=name,
        cluster_name_or_id=cluster,
        non_interactive=non_interactive,
        tail_logs=tail_logs,
    )
    print_obj("Workspace", workspace.to_dict())


@click.command(name="secret-group", help="create secret-group")
@click.argument("secret_group_name")
@handle_exception_wrapper
def create_secret_group(secret_group_name):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.create_secret_group(secret_group_name)
    print_obj(f"Secret Group", response)


@click.command(name="secret", help="create secret")
@click.argument("secret_group_id")
@click.argument("secret_key")
@click.argument("secret_value")
@handle_exception_wrapper
def create_secret(secret_group_id, secret_key, secret_value):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.create_secret(secret_group_id, secret_key, secret_value)
    print_obj(response["id"], response)


def get_create_command():
    create_command.add_command(create_workspace)

    if ENABLE_CLUSTER_COMMANDS:
        create_command.add_command(create_cluster)

    if ENABLE_SECRETS_COMMANDS:
        create_command.add_command(create_secret)
        create_command.add_command(create_secret_group)

    return create_command



from typing import Optional, List

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.lib.util import (
    resolve_deployment_or_error,
    resolve_service_or_error,
    resolve_deployments,
)
from servicefoundry.build.model.entity import Deployment


def get_deployment(
    name_or_id: str,
    service_name_or_id: Optional[str] = None,
    workspace_name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[str] = None,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Deployment:
    client = client or ServiceFoundryServiceClient.get_client()
    deployment, _, _, _ = resolve_deployment_or_error(
        name_or_id=name_or_id,
        service_name_or_id=service_name_or_id,
        cluster_name_or_id=cluster_name_or_id,
        workspace_name_or_id=workspace_name_or_id,
        non_interactive=non_interactive,
        client=client,
    )
    return deployment


def list_deployments(
    service_name_or_id: Optional[str] = None,
    workspace_name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[str] = None,
    all_: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> List[Deployment]:
    client = client or ServiceFoundryServiceClient.get_client()
    if all_:
        raise NotImplementedError
    else:
        service, workspace, cluster = resolve_service_or_error(
            name_or_id=service_name_or_id,
            workspace_name_or_id=workspace_name_or_id,
            cluster_name_or_id=cluster_name_or_id,
            non_interactive=non_interactive,
            client=client,
        )
        deployments = resolve_deployments(
            client=client,
            name_or_id=None,
            service_name_or_id=service,
            workspace_name_or_id=workspace,
            cluster_name_or_id=cluster,
            ignore_context=True,
        )
    return deployments



from shutil import copytree, ignore_patterns

from ..model.build_pack import BuildPack
from ..output_callback import OutputCallBack
from ..util import clean_dir, execute, create_file_from_content


class BasePackage:
    def __init__(self, build_dir, build_pack: BuildPack):
        self.build_pack = build_pack
        self.build_dir = build_dir
        self.build_path = f"{build_dir}/build"

    def clean(self):
        clean_dir(self.build_path)

    def pre_package(self, callback: OutputCallBack):
        if self.build_pack.pre_build_script:
            callback.print_header("Going to run pre build script")
            pre_build_script = f"{self.build_dir}/preBuildScript"
            create_file_from_content(
                pre_build_script, self.build_pack.pre_build_script, executable=True
            )
            cmd = [pre_build_script]
            for line in execute(cmd):
                callback.print_line(line.rstrip())

    def package(self, callback: OutputCallBack):
        if self.build_pack.ignore_patterns:
            patterns = ignore_patterns(
                *self.build_pack.ignore_patterns, f"{self.build_dir}*"
            )
        else:
            patterns = ignore_patterns(f"{self.build_dir}")
        copytree("./", self.build_path, ignore=patterns)



import json
import logging
import time
from typing import Optional

import requests
import socketio

from ..const import SERVICE_FOUNDRY_SERVER, REFRESH_ACCESS_TOKEN_IN_SEC
from ..model.session import ServiceFoundrySession
from ..session_factory import get_session
from ..util import upload_package_to_s3, request_handling
from ..exceptions import ConfigurationException

logger = logging.getLogger(__name__)

VERSION_PREFIX = "v1"


def _get_or_throw(definition, key, error_message):
    if key not in definition:
        raise ConfigurationException(error_message)
    return definition[key]


class ServiceFoundryServiceClient:
    def __init__(
        self, session: Optional[ServiceFoundrySession], host=SERVICE_FOUNDRY_SERVER
    ):
        self.host = host
        self.session = session

    @classmethod
    def get_client(cls, auth_required=True):
        # Would be ok to prefer auth token from API instead of local session file
        if auth_required:
            session = get_session()
            if session:
                return cls(session=session)
            else:
                raise Exception("Please login before running this command.")
        else:
            # TODO (chiragjn): All usages should be made to handle session=None
            return cls(session=None)

    def check_and_refresh_session(self):
        decoded = self.session.decode()
        expiry_second = decoded["exp"]
        if expiry_second - time.time() < REFRESH_ACCESS_TOKEN_IN_SEC:
            logger.info(
                f"Going to refresh the access token {expiry_second - time.time()}."
            )
            self.session.refresh_access_token()

    def _get_header(self):
        return {"Authorization": f"Bearer {self.session.access_token}"}

    def list_workspace(self):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/workspace"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def create_workspace(self, cluster_id, name):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/workspace"
        res = requests.post(
            url,
            data={"name": name, "clusterId": cluster_id},
            headers=self._get_header(),
        )
        return request_handling(res)

    def remove_workspace(self, workspace_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/workspace/{workspace_id}"
        res = requests.delete(url, headers=self._get_header())
        return request_handling(res)

    def get_workspace_by_name(self, workspace_name, cluster_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/workspace"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={"name": workspace_name, "clusterId": cluster_id},
        )
        return request_handling(res)

    def get_workspace(self, workspace_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/workspace/{workspace_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def list_service(self):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/service"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def list_service_by_workspace(self, workspace_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/workspace/{workspace_id}/services"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def remove_service(self, service_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/service/{service_id}"
        res = requests.delete(url, headers=self._get_header())
        return request_handling(res)

    def get_service(self, service_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/service/{service_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def list_deployment(self, service_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/service/{service_id}/deployments"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_deployment(self, deployment_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/deployment/{deployment_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def create_cluster(
        self, name, region, aws_account_id, cluster_name, ca_data, server_url
    ):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/cluster"
        res = requests.post(
            url,
            data={
                "id": name,
                "region": region,
                "authData": {
                    "awsAccountID": aws_account_id,
                    "clusterName": cluster_name,
                    "caData": ca_data,
                    "serverUrl": server_url,
                },
            },
            headers=self._get_header(),
        )
        return request_handling(res)

    def list_cluster(self):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/cluster"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_cluster(self, cluster_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/cluster/{cluster_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def remove_cluster(self, cluster_id):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/cluster/{cluster_id}"
        res = requests.delete(url, headers=self._get_header())
        return request_handling(res)

    def get_presigned_url(self, space_name, service_name, env):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/deployment/code-upload-url"
        res = requests.post(
            url,
            data={
                "workspaceFqn": space_name,
                "serviceName": service_name,
                "stage": env,
            },
            headers=self._get_header(),
        )
        return request_handling(res)

    # @TODO Get rid of env from backend and here
    def build_and_deploy(self, service_def, package_zip, env="default"):
        self.check_and_refresh_session()
        workspace = service_def["spec"]["workspace"]

        http_response = self.get_presigned_url(
            workspace, service_def["spec"]["name"], env
        )
        upload_package_to_s3(http_response, package_zip)

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/deployment"
        data = {
            "workspaceFqn": workspace,
            "componentDef": service_def,
            "s3Bucket": http_response["s3Bucket"],
            "s3Key": http_response["s3Key"],
        }

        deploy_response = requests.post(url, json=data, headers=self._get_header())
        return request_handling(deploy_response)

    def create_secret_group(self, name):
        self.check_and_refresh_session()
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret-group/"
        res = requests.post(url, headers=self._get_header(), json={"name": name})
        return request_handling(res)

    def delete_secret_group(self, id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret-group/{id}"
        res = requests.delete(url, headers=self._get_header())
        return request_handling(res)

    def get_secret_group(self, id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret-group/{id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def create_secret(self, secret_group_id, key, value):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret/"
        res = requests.post(
            url,
            headers=self._get_header(),
            json={"secretGroupId": secret_group_id, "key": key, "value": value},
        )
        return request_handling(res)

    def delete_secret(self, id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret/{id}"
        res = requests.delete(url, headers=self._get_header())
        return request_handling(res)

    def get_secret(self, id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret/{id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_secrets_in_group(self, secret_group_id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret/list-by-secret-group/{secret_group_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_secret_groups(self):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/secret-group/findAll"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def tail_logs(
        self, runId, callback=lambda log_line: print(log_line, end=""), wait=False
    ):
        self.check_and_refresh_session()
        sio = socketio.Client(request_timeout=60)
        callback("Waiting for the task to start...\n")

        @sio.event
        def logs(data):
            try:
                log_line = json.loads(data)["body"]["log"]
                callback(log_line)
            except Exception:
                pass

        # TODO: We are limiting the behavior to pure websocket since we were intermittently getting a 400
        # when the library was attempting to poll. Need to debug this further.
        sio.connect(SERVICE_FOUNDRY_SERVER, transports="websocket")
        sio.emit("logs", json.dumps({"runId": runId, "startTs": "0"}))
        if wait:
            sio.wait()

    def fetch_logs(self, runId):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/logs/runs/{runId}?startTs=0&endTs=5000000000&limit=1000&direction=asc"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_authorization_for_resource(self, resource_type, resource_id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/authorize/{resource_type}/{resource_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def create_authorization(self, resource_id, resource_type, user_id, role):
        # @TODO instead of user_id pass emailID. Need to be done once API is available on auth.
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/authorize"
        res = requests.post(
            url,
            headers=self._get_header(),
            json={
                "resourceId": resource_id,
                "resourceType": resource_type,
                "userName": user_id,
                "userType": "USER",
                "role": role,
            },
        )
        return request_handling(res)

    def delete_authorization(self, id):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/authorize/{id}"
        res = requests.delete(url, headers=self._get_header())
        return request_handling(res)

    def update_authorization(self, id, role):
        self.check_and_refresh_session()

        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/authorize"
        res = requests.patch(
            url, headers=self._get_header(), json={"id": id, "role": role}
        )
        return request_handling(res)

    def get_templates_list(self):
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/template"
        res = requests.get(url)
        return request_handling(res)

    def get_template_by_id(self, template_id):
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/template/{template_id}"
        res = requests.get(url)
        return request_handling(res)

    def get_base_by_id(self, base_id):
        url = f"{SERVICE_FOUNDRY_SERVER}/{VERSION_PREFIX}/template/base/{base_id}"
        res = requests.get(url)
        return request_handling(res)



import json
import os
import time

import jwt

from servicefoundry.build.const import SESSION_FILE


class ServiceFoundrySession:
    def __init__(
        self,
        client_id,
        access_token,
        refresh_token,
        session_file_location=SESSION_FILE,
        refresher=None,
        cluster=None,
        workspace=None,
    ):
        self.client_id = client_id
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.session_file_location = session_file_location
        self.refresher = refresher
        self.cluster = cluster
        self.workspace = workspace

    def save_session(self):
        with open(self.session_file_location, "w") as file:
            json.dump(
                {
                    "client_id": self.client_id,
                    "access_token": self.access_token,
                    "refresh_token": self.refresh_token,
                    "cluster": self.cluster,
                    "workspace": self.workspace,
                },
                file,
            )
            # Ensure it is written to disk
            file.flush()
            os.fsync(file.fileno())

    def refresh_access_token(self):
        self.access_token, self.refresh_token = self.refresher(self)
        self.save_session()

    def get_expiry(self):
        return self.decode()["exp"]

    def get_ttl(self):
        return self.get_expiry() - time.time()

    def get_user_details(self):
        decoded = self.decode()
        return {"username": decoded["preferred_username"], "email": decoded["email"]}

    def decode(self):
        return jwt.decode(self.access_token, options={"verify_signature": False})

    def set_cluster(self, cluster):
        self.cluster = cluster
        return cluster

    def get_cluster(self):
        return self.cluster

    def set_workspace(self, workspace):
        self.workspace = workspace
        return workspace

    def get_workspace(self):
        return self.workspace



import os
import re
import yaml
import shutil
from jsonschema.exceptions import ValidationError
from mako.template import Template

from .generated_definition import GeneratedDefinition
from .sf_definition import SFDefinition
from .template_parameters import get_template_parameter
from .util import get_template_file, load_yaml_from_file, validate_schema
from ..exceptions import ConfigurationException

SPEC = "spec"
OVERWRITES = "overwrites"


class Parameters(dict):
    def __init__(self, *args, **kwargs):
        super(Parameters, self).__init__(*args, **kwargs)
        self.__dict__ = self


class SfTemplate:
    def __init__(self, template_id, template_folder, template):
        self.template_id = template_id
        self.template_folder = template_folder
        try:
            validate_schema(template, "schema/template_schema.json")
        except ValidationError as err:
            raise ConfigurationException(f"Template validation failed. {err.message}")

        spec = template[SPEC]
        self.base_build = spec["baseBuild"]
        self.base_component = spec["baseComponent"]
        self.post_init_instruction = spec["postInitInstruction"]
        self.overwrite = spec[OVERWRITES]
        self.parameters = [
            get_template_parameter(parameter) for parameter in spec["parameters"]
        ]
        self.comments = spec["comments"]

    def generate_service_def(self, sf_definition: SFDefinition):
        parameters = Parameters(sf_definition.parameters)
        base_build_id = self.base_build
        base_component_id = self.base_component
        resolved_overrides = self._resolve_variables(self.overwrite, parameters)
        if sf_definition.overwrites:
            resolved_overrides.extend(
                self._resolve_variables(sf_definition.overwrites, parameters)
            )

        definition = GeneratedDefinition(
            base_build_id, base_component_id, resolved_overrides
        )
        return definition

    def generate_project(self, parameters):
        # Create new folder to hold template and rendered values.
        project_folder = parameters.get("service_name", f"{self.template_id}_service")

        if os.path.exists(project_folder):
            raise ConfigurationException(
                f"Failed to initialize project. Directory {project_folder} already exist."
            )
        else:
            shutil.copytree(self.template_folder, project_folder)
            os.remove(f"{project_folder}/template.yaml")
            # Write parameters into servicefoundry.yaml
            with open(f"{project_folder}/servicefoundry.yaml", "w") as template_file:
                yaml.dump(
                    {
                        "template": f"truefoundry.com/v1/{self.template_id}",
                        "parameters": parameters,
                    },
                    template_file,
                    sort_keys=False,
                )
                template_file.write("overwrites:")
                template_file.write(self.comments)
            return project_folder

    def _resolve_variables(self, overwrite_dict, parameters):
        resolved_values = []
        for overwrite_key, overwrite_value in overwrite_dict.items():
            resolved_values.append(
                (
                    overwrite_key,
                    self._resolve_variable(overwrite_key, overwrite_value, parameters),
                )
            )
        return resolved_values

    def _resolve_variable(self, key, value, parameters):
        if isinstance(value, dict):
            ret_value = {}
            for k, v in value.items():
                ret_value[k] = self._resolve_variable(key, v, parameters)
            return ret_value
        if isinstance(value, list):
            ret_value = []
            for item in value:
                ret_value.append(self._resolve_variable(key, item, parameters))
            return ret_value
        if isinstance(value, (int, float)):
            return value
        # Check if it's a simple substitution
        match = re.match("^\$\{parameters\.([A-Za-z0-9]+)\}$", value)
        if match:
            variable = match.group(1)
            if variable in parameters:
                return parameters[variable]
            else:
                raise ConfigurationException(
                    f"Failed to parse {key}. Parameters doesn't have {variable}"
                )
        try:
            template = Template(value)
            return template.render(parameters=parameters)
        except AttributeError as e:
            raise ConfigurationException(f"Failed to parse {key}. {e}")

    @classmethod
    def get_template(cls, template_name):
        split = template_name.split("/")
        if len(split) != 3:
            raise ConfigurationException(f"Incorrect template {template_name}")
        template_id = split[2]
        template_folder = get_template_file(template_id)
        template_yaml = load_yaml_from_file(f"{template_folder}/template.yaml")
        return cls(template_id, template_folder, template_yaml)



import yaml
from servicefoundry.build.exceptions import ConfigurationException

TEMPLATE = "template"
PARAMETERS = "parameters"
OVERWRITES = "overwrites"


class SFDefinition:
    def __init__(self, definition, file_name):
        if TEMPLATE not in definition:
            raise ConfigurationException(
                f"File {file_name} has missing field '{TEMPLATE}'."
            )
        self.template = definition[TEMPLATE]
        if PARAMETERS not in definition:
            raise ConfigurationException(
                f"File {file_name} has missing field '{PARAMETERS}'."
            )
        self.parameters = definition[PARAMETERS]
        self.overwrites = definition[OVERWRITES] if OVERWRITES in definition else None

    @classmethod
    def create(cls, file):
        with open(file) as f:
            try:
                service_foundry = yaml.full_load(f)
            except yaml.YAMLError as exc:
                raise ConfigurationException(f"File {file} is not in yaml format.")
            return cls(service_foundry, file)



import os
from pathlib import Path

SERVICE_FOUNDRY_SERVER = (
    os.getenv("SFY_SERVER")
    or "https://sf-server.tfy-ctl-euwe1-devtest.devtest.truefoundry.tech"
)

# Auth related config
# TODO: Call service foundry to get this.
DEFAULT_TENANT_ID = os.getenv("SFY_TENANT_ID") or "895253af-ec9d-4be6-83d1-6f248e644e79"
AUTH_UI = os.getenv("SFY_AUTH_UI") or "https://app.devtest.truefoundry.tech"
AUTH_SERVER = (
    os.getenv("SFY_AUTH_SERVER")
    or "https://auth-server.tfy-ctl-euwe1-devtest.devtest.truefoundry.tech"
)
SESSION_FILE = str(Path.home() / ".truefoundry")

# Build related Config
SERVICE_DEF_FILE_NAME = "servicefoundry.yaml"
BUILD_DIR = ".servicefoundry"

COMPONENT = "Component"
BUILD_PACK = "BuildPack"
KIND = "kind"

# Polling during login redirect
MAX_POLLING_RETRY = 100
POLLING_SLEEP_TIME_IN_SEC = 4

# Refresh access token cutoff
REFRESH_ACCESS_TOKEN_IN_SEC = 10 * 60



from fastapi import FastAPI

from .metrics_router import metrics_router
from .prometheus_middleware import PrometheusMiddleware


def fast_api():
    app = FastAPI()
    app.add_middleware(PrometheusMiddleware)
    app.add_route("/metrics", metrics_router)

    return app



import logging
import re
import pkg_resources
from pkg_resources import Requirement

from .package_docker import PackageDocker
from ..output_callback import OutputCallBack
from ..util import read_text_from_file, create_file_from_content, manage_file_diff

logger = logging.getLogger()

NEWLINE = "\n"
REQUIREMENTS_TXT = "requirements.txt"
MANAGED_MESSAGE = """
# ServicefoundryManaged
# Below this line dependencies are managed by servicefoundry.
# Any package mentioned after this will get auto updated from installed packages.
# If you install a new package, that would also be auto added.
# If you uninstall a package, u have to remove them manually.
""".splitlines()

NOT_INSTALLED_MESSAGE = """
# Below package are not installed. You can safely remove these.

""".splitlines()


class SfPackage:
    def __init__(self, package_name, package_version, is_installed=False):
        self.package_name = package_name
        self.package_version = package_version
        self.is_installed = is_installed

    def update_version(self, version):
        self.is_installed = True
        self.package_version = version

    def to_str(self):
        return f"{self.package_name}=={self.package_version}"


class PackagePython(PackageDocker):
    def package(self, callback: OutputCallBack):
        if self.build_pack.dependency.auto_update:
            source_lines = read_text_from_file(REQUIREMENTS_TXT)
            target_lines = self._update_requirements_txt(source_lines)
            manage_file_diff(source_lines, target_lines, REQUIREMENTS_TXT, callback)
            create_file_from_content(REQUIREMENTS_TXT, NEWLINE.join(target_lines))

        super().package(callback)

    def _update_requirements_txt(self, source_lines):
        user_lines = []
        user_packages = set()
        sf_packages = {}
        is_managed_block = False

        # Parse source lines
        for line in source_lines:
            if re.match("^\s*#\s*ServicefoundryManaged\s*$", line):
                is_managed_block = True
            if not is_managed_block:
                user_lines.append(line)
            sline = line.strip()
            if sline.strip() != "" and re.match("^\s*#", line) is None:
                requirement = Requirement(line)
                package_name = requirement.key.lower()
                if not is_managed_block:
                    user_packages.add(package_name)
                else:
                    sf_packages[package_name] = SfPackage(
                        package_name, requirement.specs[0][1]
                    )

        # Add and update installed packages
        installed_packages = {
            d.project_name.lower(): d.version for d in pkg_resources.working_set
        }
        for package_name, package_version in installed_packages.items():
            if package_name not in user_packages:
                if package_name in sf_packages:
                    sf_packages[package_name].update_version(package_version)
                else:
                    sf_packages[package_name] = SfPackage(
                        package_name, package_version, is_installed=True
                    )

        # Generate lines.
        ret_lines = []
        for line in user_lines:
            ret_lines.append(line)
        ret_lines.extend(MANAGED_MESSAGE)
        for package_name, package in sorted(sf_packages.items()):
            if package.is_installed:
                ret_lines.append(package.to_str())
        ret_lines.extend(NOT_INSTALLED_MESSAGE)
        for package_name, package in sorted(sf_packages.items()):
            if not package.is_installed:
                ret_lines.append(package.to_str())
        return ret_lines



from typing import Union, List, Tuple, Any

from servicefoundry.build.const import KIND
from servicefoundry.build.exceptions import ConfigurationException

STRING = "string"
NUMBER = "number"
OPTIONS = "options"
WORKSPACE = "tfy-workspace"

default = Union[str, int, float]


class TemplateParameter:
    id: str
    kind: str
    prompt: str
    default: default

    def __init__(self, json_def):
        self.id = json_def["id"]
        self.kind = json_def["kind"]
        self.prompt = json_def["prompt"]
        if "default" in json_def:
            self.default = json_def["default"]
        else:
            self.default = ""


class OptionsTemplateParameter(TemplateParameter):
    options: List[str]

    def __init__(self, json_def):
        super().__init__(json_def)
        self.options = json_def[OPTIONS]


def _get_workspace_choices(client):
    spaces = client.list_workspace()
    space_choices = [
        (space["name"], space["fqn"])
        for space in spaces
        if space["status"] == "CREATE_SPACE_SUCCEEDED"
    ]
    return space_choices


class WorkspaceTemplateParameter(TemplateParameter):
    def __init__(self, json_def):
        super().__init__(json_def)

    def get_workspaces(self, client):
        return _get_workspace_choices(client)


def get_template_parameter(parameter):
    kind = parameter[KIND]
    if kind in [STRING, NUMBER]:
        return TemplateParameter(parameter)
    if kind == OPTIONS:
        return OptionsTemplateParameter(parameter)
    if kind == WORKSPACE:
        return WorkspaceTemplateParameter(parameter)
    raise ConfigurationException(f"Unhandled template parameter kind '{kind}'")



import logging

import rich_click as click

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.cli.display_util import print_obj
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.group(name="auth", help="servicefoundry auth list|create|update|remove")
def authorize():
    pass


@click.command(name="create", help="create auth")
@click.argument("resource_type")
@click.argument("resource_id")
@click.argument("user_id")
@click.argument("role")
@handle_exception_wrapper
def create(resource_id, resource_type, user_id, role):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.create_authorization(
        resource_id, resource_type, user_id, role
    )
    print_obj(f"Auth for {resource_type}: {resource_id}", response)


@click.command(name="update", help="update auth")
@click.argument("authorization_id")
@click.argument("role")
@handle_exception_wrapper
def update(authorization_id, role):
    tfs_client = ServiceFoundryServiceClient.get_client()
    _ = tfs_client.update_authorization(authorization_id, role)


def get_authorization_command():
    authorize.add_command(create)
    authorize.add_command(update)
    return authorize



from typing import Optional

import rich_click as click

from servicefoundry.build.clients.auth_service_client import AuthServiceClient
from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.const import DEFAULT_TENANT_ID, SESSION_FILE
from servicefoundry.build.exceptions import BadRequestException
from servicefoundry.build.lib.messages import (
    PROMPT_ALREADY_LOGGED_OUT,
    PROMPT_LOGOUT_SUCCESSFUL,
    PROMPT_LOGIN_SUCCESSFUL,
)
from servicefoundry.build.model.entity import Cluster
from servicefoundry.build.model.session import ServiceFoundrySession
from servicefoundry.build.session_factory import get_session
from servicefoundry.build.session_factory import logout_session


def _set_cluster_if_only_one():
    client = ServiceFoundryServiceClient.get_client()
    clusters = client.list_cluster()
    if len(clusters) == 1:
        cluster = Cluster.from_dict(clusters[0])
        client.session.set_cluster(cluster.to_dict_for_session())
        client.session.save_session()


def _login_using_device_code() -> ServiceFoundrySession:
    auth_client = AuthServiceClient()
    url, user_code, device_code = auth_client.get_device_code(DEFAULT_TENANT_ID)
    console.print(f"Login Code: {user_code}")
    console.print(
        f"Waiting for authentication. Go to the following url to complete the authentication: {url}"
    )
    click.launch(url)
    session = auth_client.poll_for_auth(DEFAULT_TENANT_ID, device_code)
    session.save_session()
    console.print(PROMPT_LOGIN_SUCCESSFUL)
    console.print(f"Session file stored at {SESSION_FILE}")
    return session


def _login_with_api_key(api_key: str) -> ServiceFoundrySession:
    session = AuthServiceClient().login_with_api_token(
        client_id=DEFAULT_TENANT_ID, api_key=api_key
    )
    console.print(PROMPT_LOGIN_SUCCESSFUL)
    return session


def login(api_key: Optional[str] = None, non_interactive: bool = True):
    try:
        session = get_session()
        user = session.get_user_details()
        console.print(
            f"[yellow]You are logged in as {user['username']!r} with email {user['email']!r}[/]",
        )
    except BadRequestException:
        if non_interactive:
            if not api_key:
                raise ValueError("`api_key` is required in non interactive mode")
            _login_with_api_key(api_key=api_key)
        else:
            _login_using_device_code()
        # TODO (chiragjn): Done to have user make one less choice till we only have one cluster on backend
        _set_cluster_if_only_one()


def logout(non_interactive: bool = True):
    try:
        logout_session()
    except BadRequestException:
        console.print(PROMPT_ALREADY_LOGGED_OUT)
    else:
        console.print(PROMPT_LOGOUT_SUCCESSFUL)



import yaml
from pathlib import Path
from jsonschema.exceptions import ValidationError

from .util import validate_schema, load_yaml_from_file
from ..clients.service_foundry_client import ServiceFoundryServiceClient
from ..const import BUILD_PACK, COMPONENT, BUILD_DIR
from ..model.build_pack import BuildPack
from ..util import download_file
from ..exceptions import ConfigurationException

SPEC = "spec"


def get_base_yaml(base_id):
    base_file = f"{BUILD_DIR}/template/{base_id}"

    if not Path(base_file).is_file():
        tfs_client = ServiceFoundryServiceClient.get_client(auth_required=False)
        package_url = tfs_client.get_base_by_id(base_id)["url"]
        download_file(package_url, base_file)

    return load_yaml_from_file(base_file)


class GeneratedDefinition:
    def __init__(self, base_build_id, base_component_id, resolved_overrides):
        self.base_def = {
            BUILD_PACK: get_base_yaml(base_build_id),
            COMPONENT: get_base_yaml(base_component_id),
        }
        for key, value in resolved_overrides:
            self._apply_overwrite(self.base_def, key.split("."), value)

    def write(self, file="servicefoundry.lock.yaml"):
        with open(file, "w") as outfile:
            yaml.dump_all(self.base_def.values(), outfile, default_flow_style=False)

    def get_build_pack(self):
        return BuildPack(**self.base_def[BUILD_PACK][SPEC])

    def get_component(self):
        return self.base_def[COMPONENT]

    def validate(self):
        try:
            validate_schema(self.base_def[BUILD_PACK], "schema/build_pack_schema.json")
        except ValidationError as err:
            raise ConfigurationException(f"Build pack validation failed. {err.message}")

        try:
            validate_schema(self.base_def[COMPONENT], "schema/component_schema.json")
        except ValidationError as err:
            raise ConfigurationException(f"Component validation failed. {err.message}")

    def _apply_overwrite(self, definition, keys, value):
        key = keys[0]
        if len(keys) == 1:
            definition[keys[0]] = value
            return
        if key in definition:
            self._apply_overwrite(definition[key], keys[1:], value)
        else:
            raise ConfigurationException(f"{key} not found in {definition}")



import difflib
import logging
import os
import shutil
import subprocess
import tarfile

import pkg_resources
import requests

from .exceptions import BadRequestException
from .output_callback import OutputCallBack

logger = logging.getLogger()


def read_text(file_name, name=__name__):
    return pkg_resources.resource_string(name, file_name).decode("utf-8")


def read_text_from_file(file_path):
    with open(file_path) as file:
        return [line.rstrip() for line in file.readlines()]


def clean_dir(dir_name):
    if os.path.isfile(dir_name):
        os.remove(dir_name)
    if os.path.isdir(dir_name):
        shutil.rmtree(dir_name)


def make_executable(path):
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2
    os.chmod(path, mode)


def create_file_from_content(file_name, content, executable=False):
    with open(file_name, "w") as text_file:
        text_file.write(content)
    if executable:
        make_executable(file_name)


def upload_package_to_s3(metadata, package_file):
    with open(package_file, "rb") as file_to_upload:
        http_response = requests.put(metadata["url"], data=file_to_upload)

        if http_response.status_code not in [204, 201, 200]:
            raise RuntimeError(f"Failed to upload to S3 {http_response.content}")


def request_handling(res):
    try:
        status_code = res.status_code
    except Exception:
        raise Exception("Unknown error occurred. Couldn't get status code.")
    if 200 <= status_code <= 299:
        if res.content == b"":
            return None
        return res.json()
    if 400 <= status_code <= 499:
        try:
            message = res.json()["message"]
        except Exception:
            message = res
        raise BadRequestException(res.status_code, message)
    if 500 <= status_code <= 599:
        raise Exception(res.content)


def run_process(cmd):
    return subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
    )


def execute(cmd):
    popen = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
    )
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    for stderr_line in iter(popen.stderr.readline, ""):
        yield stderr_line
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, " ".join(cmd))


def make_tarfile(output_filename, source_dir):
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname=os.path.basename(source_dir))


def download_file(url, file_path):
    r = requests.get(url, allow_redirects=True)
    open(file_path, "wb").write(r.content)


def uncompress_tarfile(file_path, destination):
    file = tarfile.open(file_path)
    file.extractall(destination)
    file.close()


def get_file_diff(from_lines, to_lines, from_file="Original", to_file="Current"):
    return [
        line
        for line in difflib.unified_diff(
            from_lines, to_lines, fromfile=from_file, tofile=to_file, lineterm=""
        )
    ]


def manage_file_diff(source_lines, target_lines, entity, callback: OutputCallBack):
    diffs = get_file_diff(
        source_lines, target_lines, f"Original {entity}", f"Updated {entity}"
    )

    if len(diffs) == 0:
        callback.print_line(f"No new changes are required in {entity}.")
    else:
        callback.print_line(f"Going to apply below changes in {entity}")
        callback.print_lines_in_panel(diffs, f"{entity} Diff")



class BuildPack:
    def __init__(
        self,
        type,
        ignorePatterns,
        docker,
        dependency=None,
        preBuildScript=None,
        local=None,
        **kwargs,
    ):
        self.type = type
        self.ignore_patterns = ignorePatterns
        self.docker = Docker(**docker)
        if dependency:
            self.dependency = Dependency(**dependency)
        self.pre_build_script = preBuildScript
        if local:
            self.local = Local(**local)


class Local:
    def __init__(self, requirementFiles, runCommand, **kwargs):
        self.run_command = runCommand
        self.requirement_files = requirementFiles


class Dependency:
    def __init__(self, autoUpdate, **kwargs):
        self.auto_update = autoUpdate


class Docker:
    def __init__(self, fileName, overwrite, dockerFileContent=None, **kwargs):
        self.docker_file_content = dockerFileContent
        self.file_name = fileName
        self.overwrite = overwrite



import logging
import time

import requests

from servicefoundry.build.const import (
    AUTH_SERVER,
    MAX_POLLING_RETRY,
    POLLING_SLEEP_TIME_IN_SEC,
    AUTH_UI,
)
from servicefoundry.build.model.session import ServiceFoundrySession
from servicefoundry.build.util import request_handling

logger = logging.getLogger(__name__)

TRUE_FOUNDRY_AUTH_TOKEN = "TRUE_FOUNDRY_AUTH_TOKEN"


class AuthServiceClient:
    def __init__(self, host=AUTH_SERVER, auth_ui=AUTH_UI):
        self.host = host
        self.auth_ui = auth_ui

    def refresh_token(self, session: ServiceFoundrySession):
        url = f"{self.host}/api/v1/oauth/token/refresh"
        data = {"clientId": session.client_id, "refreshToken": session.refresh_token}
        res = requests.post(url, data=data)
        res = request_handling(res)
        return res["accessToken"], res["refreshToken"]

    def get_device_code(self, client_id):
        url = f"{self.host}/api/v1/oauth/device"
        data = {"clientId": client_id}
        res = requests.post(url, data=data)
        res = request_handling(res)
        return (
            f"{self.auth_ui}/authorize/device?userCode={res['userCode']}",
            res["userCode"],
            res["deviceCode"],
        )

    def poll_for_auth(self, tenant_id, device_code):
        i = 0
        while i < MAX_POLLING_RETRY:
            try:
                return self._poll_for_auth(tenant_id, device_code)
            except RetryError:
                time.sleep(POLLING_SLEEP_TIME_IN_SEC)
                i = i + 1

    def _poll_for_auth(self, tenant_id, device_code):
        url = f"{self.host}/api/v1/oauth/device/token"
        data = {"clientId": tenant_id, "deviceCode": device_code}
        res = requests.post(url, data=data)
        res = request_handling(res)
        if "message" in res and res["message"] == "authorization_pending":
            raise RetryError()
        return ServiceFoundrySession(
            tenant_id,
            res["accessToken"],
            res["refreshToken"],
            refresher=self.refresh_token,
        )

    def login_with_api_token(self, client_id, api_key):
        url = f"{self.host}/api/v1/oauth/api-keys/token"
        data = {"clientId": client_id, "apiKey": api_key}
        res = requests.post(url, data=data)
        res = request_handling(res)
        return ServiceFoundrySession(
            client_id,
            res["accessToken"],
            res["refreshToken"],
            refresher=self.refresh_token,
        )

    def validate_user_code(self, session: ServiceFoundrySession, user_code):
        url = f"{self.host}/api/v1/oauth/device/validate"
        data = {
            "clientId": session.client_id,
            "refreshToken": session.refresh_token,
            "userCode": user_code,
        }
        res = requests.post(url, data=data)
        res = request_handling(res)
        return ServiceFoundrySession(
            session.client_id,
            res["accessToken"],
            res["refreshToken"],
            self.refresh_token,
        )


class RetryError(Exception):
    def __init__(self):
        super()



import json
import os

from servicefoundry.build.clients.auth_service_client import AuthServiceClient
from servicefoundry.build.model.session import ServiceFoundrySession
from .const import SESSION_FILE, DEFAULT_TENANT_ID
from .exceptions import BadRequestException


def get_session(session_file=SESSION_FILE):
    if os.getenv("SERVICE_FOUNDRY_API_KEY"):
        auth_client = AuthServiceClient()
        return auth_client.login_with_api_token(
            DEFAULT_TENANT_ID, os.getenv("SERVICE_FOUNDRY_API_KEY")
        )

    if os.path.isfile(session_file):
        with open(session_file, "r") as file:
            data = json.load(file)
            return ServiceFoundrySession(
                **data, refresher=AuthServiceClient().refresh_token
            )
    else:
        raise BadRequestException(403, f"Please login before running this command.")


def logout_session(session_file=SESSION_FILE):
    # TODO: Implement logout if using api key
    if os.path.isfile(session_file):
        with open(session_file, "r") as file:
            data = json.load(file)
            _ = ServiceFoundrySession(
                **data, refresher=AuthServiceClient().refresh_token
            )
        os.remove(session_file)
    else:
        raise BadRequestException(403, f"Please login before running this command.")



import dataclasses
from typing import Dict, Any, List, ClassVar


# TODO: Hide id fields from display columns
# TODO: make sure all entities have createdAt and updatedAt
# TODO: switch to datetime.datetime for createdAt and updatedAt
# TODO: switch to Enums for str literals
# TODO: Implement NotImplementedError sections


@dataclasses.dataclass
class Entity:
    display_columns: ClassVar[List[str]] = []

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]):
        known_attr_names = {field.name for field in dataclasses.fields(cls)}
        known_attrs = {}
        extra_attrs = {}
        for k, v in dct.items():
            if k in known_attr_names:
                known_attrs[k] = v
            else:
                extra_attrs[k] = v

        instance = cls(**known_attrs)
        for attr_name, attr_value in extra_attrs.items():
            if not hasattr(instance, attr_name):
                setattr(instance, attr_name, attr_value)
        return instance

    def to_dict(self) -> Dict[str, Any]:
        return vars(self).copy()


@dataclasses.dataclass
class Cluster(Entity):
    id: str
    name: str
    fqn: str
    region: str
    createdAt: str
    updatedAt: str
    display_columns: ClassVar[List[str]] = [
        # "id",
        "name",
        "fqn",
        "region",
        "createdAt",
        "updatedAt",
    ]

    def to_dict_for_session(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @property
    def workspaces(self) -> List["Workspace"]:
        raise NotImplementedError


@dataclasses.dataclass
class Workspace(Entity):
    id: str
    fqn: str
    name: str
    clusterId: str
    createdBy: str
    status: str
    createdAt: str
    updatedAt: str
    display_columns: ClassVar[List[str]] = [
        # "id",
        "name",
        "createdBy",
        "status",
        "createdAt",
    ]

    def to_dict_for_session(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @property
    def cluster(self) -> Cluster:
        raise NotImplementedError

    @property
    def services(self) -> List["Service"]:
        raise NotImplementedError


@dataclasses.dataclass
class Service(Entity):
    id: str
    name: str
    fqn: str
    workspaceId: str
    status: str
    metadata: Dict[str, Any]  # TODO: flatten if needed
    display_columns: ClassVar[List[str]] = [
        # "id",
        "name",
        "status",
        "metadata",
    ]

    @property
    def workspace(self) -> Workspace:
        raise NotImplementedError

    @property
    def deployments(self) -> List["Deployment"]:
        raise NotImplementedError


@dataclasses.dataclass
class Deployment(Entity):
    id: str
    name: str
    fqn: str
    serviceId: str
    status: int
    createdBy: str
    display_columns: ClassVar[List[str]] = [
        # "id",
        "name",
        "serviceId",
        "createdBy",
        "status",
    ]

    @property
    def service(self) -> Service:
        raise NotImplementedError


# TODO: Should treat displaying and handling these with more respect as it is sensitive data


@dataclasses.dataclass
class Secret(Entity):
    id: str
    name: str
    secretGroupId: str
    value: str
    fqn: str
    createdBy: str
    createdAt: str
    updatedAt: str
    display_columns: ClassVar[List[str]] = [
        "id",
        "name",
        "secretGroupId",
        "createdAt",
        "updatedAt",
    ]

    @property
    def secret_group(self) -> "SecretGroup":
        raise NotImplementedError


@dataclasses.dataclass
class SecretGroup(Entity):
    id: str
    fqn: str
    name: str
    createdBy: str
    createdAt: str
    updatedAt: str
    associatedSecrets: List[Secret]
    display_columns: ClassVar[List[str]] = ["id", "name", "createdAt", "updatedAt"]

    @classmethod
    def from_dict(cls, dct):
        dct["associatedSecrets"] = [
            Secret.from_dict(s) for s in dct.get("associatedSecrets", [])
        ]
        super().from_dict(dct)



import os

from prometheus_client import (
    CONTENT_TYPE_LATEST,
    REGISTRY,
    CollectorRegistry,
    generate_latest,
)
from prometheus_client.multiprocess import MultiProcessCollector
from starlette.requests import Request
from starlette.responses import Response


def metrics_router(request: Request) -> Response:
    if "prometheus_multiproc_dir" in os.environ:
        registry = CollectorRegistry()
        MultiProcessCollector(registry)
    else:
        registry = REGISTRY

    return Response(
        generate_latest(registry), headers={"Content-Type": CONTENT_TYPE_LATEST}
    )



from typing import Optional, Dict, Any, List

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib.messages import (
    PROMPT_REMOVING_SERVICE,
    PROMPT_USING_WORKSPACE_CONTEXT,
)
from servicefoundry.build.lib.util import (
    resolve_services,
    resolve_service_or_error,
    resolve_workspace_or_error,
)
from servicefoundry.build.model.entity import Service


def get_service(
    name_or_id: str,
    workspace_name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[str] = None,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Service:
    client = client or ServiceFoundryServiceClient.get_client()
    service, _, _ = resolve_service_or_error(
        name_or_id=name_or_id,
        workspace_name_or_id=workspace_name_or_id,
        cluster_name_or_id=cluster_name_or_id,
        non_interactive=non_interactive,
        client=client,
    )
    return service


def list_services(
    workspace_name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[str] = None,
    all_: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> List[Service]:
    client = client or ServiceFoundryServiceClient.get_client()
    if all_:
        services = resolve_services(
            client=client,
            name_or_id=None,
            workspace_name_or_id=workspace_name_or_id,
            cluster_name_or_id=None,
            ignore_context=True,
        )
    else:
        workspace, cluster = resolve_workspace_or_error(
            name_or_id=workspace_name_or_id,
            cluster_name_or_id=cluster_name_or_id,
            non_interactive=non_interactive,
            client=client,
        )
        console.print(PROMPT_USING_WORKSPACE_CONTEXT.format(workspace.name))
        services = resolve_services(
            client=client,
            name_or_id=None,
            workspace_name_or_id=workspace,
            cluster_name_or_id=cluster,
            ignore_context=True,
        )
    return services


def remove_service(
    name_or_id: str,
    workspace_name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[str] = None,
    tail_logs: bool = True,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Dict[str, Any]:
    client = client or ServiceFoundryServiceClient.get_client()
    service = get_service(
        name_or_id=name_or_id,
        workspace_name_or_id=workspace_name_or_id,
        cluster_name_or_id=cluster_name_or_id,
        non_interactive=non_interactive,
        client=client,
    )
    with console.status(PROMPT_REMOVING_SERVICE.format(service.name), spinner="dots"):
        response = client.remove_service(service.id)
        if tail_logs:
            client.tail_logs(response["runId"])
    return response



from .sf_definition import SFDefinition
from .sf_template import SfTemplate


def parse(file):
    sf_definition = SFDefinition.create(file)
    template = SfTemplate.get_template(sf_definition.template)
    out = template.generate_service_def(sf_definition)
    return out



from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.panel import Panel

from servicefoundry.cli.const import MAX_WIDTH


class RichOutputCallBack:
    console = Console(width=MAX_WIDTH)
    highlighter = ReprHighlighter()

    def print_header(self, line):
        self.console.rule(f"{line}", style="cyan")

    def print_line(self, line):
        self.console.print(line)

    def print_lines_in_panel(self, lines, header):
        self.console.print(Panel(self.highlighter("\n".join(lines)), title=header))



import os

from .const import SERVICE_DEF_FILE_NAME, BUILD_DIR
from .deploy.local_deploy import deploy as local_deploy
from .deploy.remote_deploy import deploy as remote_deploy
from .output_callback import OutputCallBack
from .package.packaging_factory import package
from .parser.parser import parse
from .exceptions import ConfigurationException

LOCAL = "local"
REMOTE = "remote"


def build_and_deploy(
    base_dir="./",
    service_def_file_name=SERVICE_DEF_FILE_NAME,
    build=REMOTE,
    callback=OutputCallBack(),
):
    current_dir = os.getcwd()
    if not base_dir.endswith("/"):
        base_dir = f"{base_dir}/"
    # We will set it back to olf directory once done.
    os.chdir(base_dir)

    try:
        if not os.path.isfile(service_def_file_name):
            raise ConfigurationException(
                f"Service definition {service_def_file_name} doesn't exist in {base_dir}"
            )

        service_defs = parse(service_def_file_name)
        service_defs.write()

        build_pack = service_defs.get_build_pack()
        component = service_defs.get_component()

        callback.print_header("Packaging")
        build_dir = BUILD_DIR
        package_dir = package(build_dir, build_pack, callback)

        if build == LOCAL:
            return local_deploy(build_pack, component, package_dir, build_dir, callback)
        elif build == REMOTE:
            return remote_deploy(component, package_dir, build_dir, callback)
        else:
            raise RuntimeError(f"Unrecognised build type {build}")
    finally:
        os.chdir(current_dir)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['servicefoundry',
 'servicefoundry.build',
 'servicefoundry.build.clients',
 'servicefoundry.build.deploy',
 'servicefoundry.build.lib',
 'servicefoundry.build.model',
 'servicefoundry.build.package',
 'servicefoundry.build.parser',
 'servicefoundry.cli',
 'servicefoundry.cli.commands',
 'servicefoundry.notebook']

package_data = \
{'': ['*'], 'servicefoundry.build': ['schema/*']}

install_requires = \
['Mako>=1.1.6,<2.0.0',
 'PyJWT>=2.3.0,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'click>=8.0.4,<9.0.0',
 'fastapi>=0.75.0,<0.76.0',
 'importlib-metadata>=4.2,<5.0',
 'importlib-resources>=5.2.0,<6.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'mistune>=0.8.4,<0.9.0',
 'prometheus_client>=0.13.1,<0.14.0',
 'python-socketio[client]>=5.5.2,<6.0.0',
 'questionary>=1.10.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich-click>=1.2.1,<2.0.0',
 'rich>=12.0.0,<13.0.0']

entry_points = \
{'console_scripts': ['servicefoundry = servicefoundry.__main__:main',
                     'sfy = servicefoundry.__main__:main']}

setup_kwargs = {
    'name': 'servicefoundry',
    'version': '0.1.23',
    'description': 'Generate deployed services from code',
    'long_description': '# Dev Setup\n\n**Setup a virtual env (or use an existing one)**\n\n```shell\ncd ~/   # or someplace else\npython3.9 -m venv venv39\n```\n\n> prefer using `python3.9-intel` on Apple Silicon\n\n**Activate the virtual environment**\n\n```shell\nsource ~/venv39/bin/activate\n```\n\n**Install `poetry` and `pre-commit`**\n\n```shell\npip install poetry pre-commit\n```\n\n**Git clone**\n\n```shell\ngit clone https://github.com/truefoundry/servicefoundry-cli\n```\n\n**Install pre-commit hooks**\n\n```shell\ncd servicefoundry-cli/\npre-commit install --install-hooks\n```\n\n> You can run the hook manually with `pre-commit run --all-files`\n\n\n\nIt is recommended to go through [Poetry Docs](https://python-poetry.org/docs/) to get a hang of development, testing and\npackaging process\n\n# Local Installation\n\n```shell\npoetry install\n```\n\n# Testing\n\n```shell\npoetry run pytest\n```\n\n# Build distributions\n\n```shell\npoetry build\n```\n\n# Usage :\n\n- Login : `servicefoundry login`\n\n- Init : `servicefoundry init`\n\n- Run : `servicefoundry run`\n\n# Process flow to use the commands\n\n- servicefoundry (should hint to servicefoundry login)\n- servicefoundry login (should hint to init, run from example folder)\n- servicefoundry init (should hint to run)\n- servicefoundry run (should hint to grafanaendpoint and servicefoundry get service <service_id>)\n\n- servicefoundry list workspace\n- servicefoundry list\n',
    'author': 'Abhishek Choudhary',
    'author_email': 'abhichoudhary06@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/innoavator/servicefoundry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from .package_docker import PackageDocker
from .package_python import PackagePython
from ..model.build_pack import BuildPack
from ..exceptions import ConfigurationException
from ..output_callback import OutputCallBack


def package(build_dir, build_pack: BuildPack, callback: OutputCallBack):
    if build_pack.type == "python":
        package = PackagePython(build_dir, build_pack)
    elif build_pack.type == "docker":
        package = PackageDocker(build_dir, build_pack)
    else:
        raise ConfigurationException(f"{build_pack.type} not supported.")
    package.clean()
    package.pre_package(callback)
    package.package(callback)
    return package.build_path



import logging
from types import SimpleNamespace

import questionary
import rich_click as click
from questionary import Choice

from servicefoundry.build import lib
from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib import workspace as workspace_lib
from servicefoundry.build.lib.util import resolve_cluster_or_error, resolve_workspaces
from servicefoundry.build.parser.sf_template import SfTemplate
from servicefoundry.build.parser.template_parameters import (
    NUMBER,
    OPTIONS,
    STRING,
    WORKSPACE,
)
from servicefoundry.build.session_factory import get_session
from servicefoundry.build.util import BadRequestException
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)

MSG_CREATE_NEW_SPACE = "Create a new workspace"


@click.command(help="Initialize new service for servicefoundry")
@handle_exception_wrapper
def init():
    # Get SFSClient
    tfs_client = ServiceFoundryServiceClient.get_client()

    # Get Session else do login
    try:
        get_session()
    except BadRequestException:
        do_login = questionary.select(
            "You need to login to create a service", ["Login", "Exit"]
        ).ask()
        if do_login == "Login":
            lib.login(non_interactive=False)
        else:
            return

    # Static call to get list of templates
    templates = tfs_client.get_templates_list()

    # Choose a template of service to be created.
    template_choices = [
        Choice(f'{t["id"]} - {t["description"]}', value=t["id"]) for t in templates
    ]
    template_id = questionary.select("Choose a template", template_choices).ask()
    sf_template = SfTemplate.get_template(f"truefoundry.com/v1/{template_id}")

    parameters = {}
    for param in sf_template.parameters:
        if param.kind == STRING:
            parameters[param.id] = questionary.text(
                param.prompt, default=param.default
            ).ask()
        elif param.kind == NUMBER:
            while True:
                value = questionary.text(param.prompt, default=str(param.default)).ask()
                if value.isdigit():
                    parameters[param.id] = int(value)
                    break
                else:
                    print("Not an integer Value. Try again")
        elif param.kind == OPTIONS:
            parameters[param.id] = questionary.select(
                param.prompt, choices=param.options
            ).ask()
        elif param.kind == WORKSPACE:
            # TODO: use get spaces
            workspaces = resolve_workspaces(
                client=tfs_client,
                name_or_id=None,
                cluster_name_or_id=None,
                ignore_context=False,
            )
            # TODO (chiragjn): should display fqn here for same workspace name across clusters in case
            #                  cluster is not set in context
            workspace_choices = [
                Choice(title=w.name, value=w)
                for w in workspaces
                if w.status == "CREATE_SPACE_SUCCEEDED"
            ]
            workspace_choices.append(
                Choice(title=MSG_CREATE_NEW_SPACE, value=MSG_CREATE_NEW_SPACE)
            )
            workspace = questionary.select(
                param.prompt, choices=workspace_choices
            ).ask()

            if workspace == MSG_CREATE_NEW_SPACE:
                cluster = resolve_cluster_or_error(
                    name_or_id=None,
                    ignore_context=False,
                    non_interactive=False,
                    client=tfs_client,
                )
                new_space_name = questionary.text(
                    "Please provide a name for your workspace"
                ).ask()
                workspace = workspace_lib.create_workspace(
                    name=new_space_name,
                    cluster_name_or_id=cluster.id,
                    tail_logs=True,
                    non_interactive=False,
                    client=tfs_client,
                )
                console.print(
                    f"Done, created new workspace with name {new_space_name!r}"
                )
            parameters[param.id] = workspace.fqn

    sf_template.generate_project(parameters)

    if sf_template.post_init_instruction:
        console.print(
            sf_template.post_init_instruction.format(
                parameters=SimpleNamespace(**parameters)
            )
        )


def get_init_command():
    return init



import os

TEMP_FOLDER = ".servicefoundry_templates"
ENABLE_CLUSTER_COMMANDS = False
ENABLE_AUTHORIZE_COMMANDS = False
ENABLE_SECRETS_COMMANDS = False
IS_DEBUG = True if os.getenv("SFY_DEBUG") else False
MAX_WIDTH = 100



import logging
from pathlib import Path

from .base_package import BasePackage
from ..exceptions import ConfigurationException
from ..output_callback import OutputCallBack
from ..util import create_file_from_content, read_text_from_file, manage_file_diff

logger = logging.getLogger()


class PackageDocker(BasePackage):
    def package(self, callback: OutputCallBack):
        if (
            Path(self.build_pack.docker.file_name).exists() is False
            or self.build_pack.docker.overwrite
        ):
            if Path(self.build_pack.docker.file_name).is_dir():
                raise ConfigurationException(
                    f"Can't overwrite Dockerfile. "
                    f"Since {self.build_pack.docker.file_name} is a directory."
                )
            source_lines = []
            if Path(self.build_pack.docker.file_name).is_file():
                source_lines = read_text_from_file(self.build_pack.docker.file_name)

            target_lines = self.build_pack.docker.docker_file_content.splitlines()
            manage_file_diff(
                source_lines, target_lines, self.build_pack.docker.file_name, callback
            )
            create_file_from_content(
                self.build_pack.docker.file_name,
                self.build_pack.docker.docker_file_content,
            )
        super().package(callback)



import os
import tarfile

from ..clients.service_foundry_client import ServiceFoundryServiceClient
from ..session_factory import get_session


def make_tarfile(output_filename, source_dir):
    with tarfile.open(output_filename, "w:gz") as tar:
        for fn in os.listdir(source_dir):
            p = os.path.join(source_dir, fn)
            tar.add(p, arcname=fn)


def deploy(service_def, package_dir, build_dir, stdout):
    package_zip = f"{build_dir}/build.tar.gz"
    make_tarfile(package_zip, package_dir)

    session = get_session()
    tf_client = ServiceFoundryServiceClient(session)
    resp = tf_client.build_and_deploy(service_def, package_zip)
    return resp



import logging

import rich_click as click

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib import deployment as deployment_lib
from servicefoundry.build.lib import service as service_lib
from servicefoundry.build.lib import workspace as workspace_lib
from servicefoundry.build.model.entity import Cluster, Secret, SecretGroup
from servicefoundry.cli.const import ENABLE_CLUSTER_COMMANDS, ENABLE_SECRETS_COMMANDS
from servicefoundry.cli.display_util import print_obj
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.group(name="get")
def get_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Get servicefoundry resources

    \b
    Supported resources:
    - workspace
    - service
    - deployment
    """
    pass


@click.command(name="cluster", help="show cluster metadata")
@click.argument("cluster_id")
@handle_exception_wrapper
def get_cluster(cluster_id):
    tfs_client = ServiceFoundryServiceClient.get_client()
    cluster = tfs_client.get_cluster(cluster_id)
    print_obj("Cluster", cluster, columns=Cluster.display_columns)


@click.command(name="workspace", help="show workspace metadata")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to find this workspace in",
)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def get_workspace(name, cluster, non_interactive):
    workspace = workspace_lib.get_workspace(
        name_or_id=name,
        cluster_name_or_id=cluster,
        non_interactive=non_interactive,
    )
    print_obj("Workspace", workspace.to_dict())


@click.command(name="service", help="show service metadata")
@click.argument("name", type=click.STRING)
@click.option(
    "-w",
    "--workspace",
    type=click.STRING,
    default=None,
    help="workspace to find this service in",
)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to find this cluster in",
)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def get_service(name, workspace, cluster, non_interactive):
    service = service_lib.get_service(
        name_or_id=name,
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        non_interactive=non_interactive,
    )
    print_obj("Service", service.to_dict())


@click.command(name="deployment", help="show deployment metadata")
@click.argument("name", type=click.STRING)
@click.option(
    "-s",
    "--service",
    type=click.STRING,
    default=None,
    help="service to which the deployment belongs to",
)
@click.option(
    "-w",
    "--workspace",
    type=click.STRING,
    default=None,
    help="workspace to find this service in",
)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to find this cluster in",
)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def get_deployment(name, service, workspace, cluster, non_interactive):
    deployment = deployment_lib.get_deployment(
        name_or_id=name,
        service_name_or_id=service,
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        non_interactive=non_interactive,
    )
    print_obj("Deployment", deployment.to_dict())


@click.command(name="secret-group", help="show secret-group")
@click.argument("secret_group_id")
@handle_exception_wrapper
def get_secret_group(secret_group_id):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.get_secret_group(secret_group_id)
    print_obj(f"Secret Group", response, columns=SecretGroup.display_columns)


@click.command(name="secret", help="show secret")
@click.argument("secret_id")
@handle_exception_wrapper
def get_secret(secret_id):
    tfs_client = ServiceFoundryServiceClient.get_client()
    response = tfs_client.get_secret(secret_id)
    print_obj(response["id"], response, columns=Secret.display_columns)


@click.command(name="context", help="show current context")
@handle_exception_wrapper
def get_current_context():
    tfs_client = ServiceFoundryServiceClient.get_client()
    cluster = tfs_client.session.get_cluster()
    workspace = tfs_client.session.get_workspace()
    if workspace:
        console.print(f"Workspace: {workspace['name']} ({cluster['id']})")
    elif cluster:
        console.print(f"No workspaces set as default ({cluster['id']})")
    else:
        console.print(
            f"Context not set. Please use `sfy use workspace` to pick a default workspace"
        )


def get_get_command():
    get_command.add_command(get_workspace)
    get_command.add_command(get_service)
    get_command.add_command(get_deployment)
    get_command.add_command(get_current_context)

    if ENABLE_CLUSTER_COMMANDS:
        get_command.add_command(get_cluster)

    if ENABLE_SECRETS_COMMANDS:
        get_command.add_command(get_secret)
        get_command.add_command(get_secret_group)

    return get_command



from threading import Thread

from IPython.core.display import display
from ipywidgets import Output, widgets

from servicefoundry.build.build import LOCAL, REMOTE, build_and_deploy
from servicefoundry.notebook.notebook_callback import NotebookOutputCallBack

thread = None


def deploy(project_folder, is_local=True):
    global thread
    output = Output(
        layout=widgets.Layout(
            width="100%", height="auto", max_height="200px", overflow="hidden scroll"
        )
    )

    if thread is not None and thread.is_alive():
        output.append_stdout("Stopping the old process.")
        thread.stop()
        thread.join()

    deployment = build_and_deploy(
        env="test",
        base_dir=project_folder,
        build=LOCAL if is_local else REMOTE,
        callback=NotebookOutputCallBack(output),
    )
    if isinstance(deployment, Thread):
        thread = deployment

    box = widgets.Box(
        children=[output], layout=widgets.Layout(width="100%", height="auto")
    )
    display(box)



from ipywidgets import Output


class NotebookOutputCallBack:
    output: Output

    def __init__(self, output):
        self.output = output

    def print_header(self, line):
        self.output.append_stdout(f"{line} ----------------------------------------\n")

    def print_line(self, line):
        self.output.append_stdout(f"{line}\n")

    def print_lines_in_panel(self, lines, header):
        self.print_header(header)
        for line in lines:
            self.print_line(line)



import json
import yaml
from pathlib import Path
from jsonschema import validate

from ..clients.service_foundry_client import ServiceFoundryServiceClient
from ..const import BUILD_DIR
from ..exceptions import ConfigurationException
from ..util import read_text, download_file, uncompress_tarfile


def validate_schema(template, schema):
    schema = read_text(schema)
    schema = json.loads(schema)
    validate(instance=template, schema=schema)


def get_template_file(template_id):
    Path(f"{BUILD_DIR}/template").mkdir(parents=True, exist_ok=True)
    template_folder = f"{BUILD_DIR}/template/{template_id}"

    if not Path(template_folder).is_dir():
        tfs_client = ServiceFoundryServiceClient.get_client(auth_required=False)
        package_url = tfs_client.get_template_by_id(template_id)["url"]
        package_file = f"{BUILD_DIR}/template/{template_id}.tgz"
        download_file(package_url, package_file)
        uncompress_tarfile(package_file, template_folder)
    return template_folder


def load_yaml_from_file(file):
    with open(file) as f:
        try:
            return yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise ConfigurationException(f"File {file} is not in yaml format. {exc}")



import rich_click as click

from servicefoundry.cli.version import __version__

from .config import CliConfig
from .const import MAX_WIDTH

click.rich_click.STYLE_ERRORS_SUGGESTION = "blue italic"
click.rich_click.MAX_WIDTH = MAX_WIDTH
click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.USE_RICH_MARKUP = True
click.rich_click.STYLE_HELPTEXT = ""

from servicefoundry.cli.commands import (
    get_create_command,
    get_deploy_command,
    get_get_command,
    get_init_command,
    get_list_command,
    get_login_command,
    get_logout_command,
    get_remove_command,
    get_set_command,
)


def create_service_foundry_cli():
    """Generates CLI by combining all subcommands into a main CLI and returns in
    Returns:
        function: main CLI functions will all added sub-commands
    """
    _cli = service_foundry_cli
    _cli.add_command(get_init_command())
    _cli.add_command(get_deploy_command())
    _cli.add_command(get_login_command())
    _cli.add_command(get_get_command())
    _cli.add_command(get_list_command())
    _cli.add_command(get_remove_command())
    _cli.add_command(get_create_command())
    _cli.add_command(get_logout_command())
    _cli.add_command(get_set_command())
    return _cli


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("--json", is_flag=True)
@click.option("--logs", is_flag=True)
@click.version_option(__version__)
def service_foundry_cli(json, logs):
    """
    Servicefoundry provides an easy way to deploy your code as a web service.
    \b

    [yellow]To start, login to your truefoundry account with [bold]sfy login[/][/]

    \b
    [yellow]Once logged in, start a new service with [bold]sfy init[/][/]
    """
    # TODO (chiragjn): Change this to -o json|yaml|table|pager
    CliConfig.set("json", json)
    CliConfig.set("logs", logs)



from typing import Optional, List, Union, Tuple, TypeVar

import questionary

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.console import console
from servicefoundry.build.lib.messages import (
    PROMPT_USING_CLUSTER_CONTEXT,
    PROMPT_USING_WORKSPACE_CONTEXT,
)
from servicefoundry.build.model.entity import Cluster, Workspace, Service, Deployment

# TODO: Move type casting downwards into `ServiceFoundrySession` and `ServiceFoundryServiceClient`
# TODO: Abstract duplicated code across resolving different entities
T = TypeVar("T")


def _filter_by_name_or_id(instances: List[T], name_or_id: Optional[str]) -> List[T]:
    if name_or_id:
        found = [i for i in instances if i.name == name_or_id]
        if not found:
            found = [i for i in instances if i.id == name_or_id]
    else:
        found = instances
    return found


def resolve_clusters(
    client: ServiceFoundryServiceClient,
    name_or_id: Optional[str] = None,
    ignore_context: bool = False,
) -> List[Cluster]:
    if not ignore_context and not name_or_id:
        cluster = client.session.get_cluster()
        if cluster:
            cluster = Cluster.from_dict(cluster)
            return [cluster]
    clusters = [Cluster.from_dict(c) for c in client.list_cluster()]
    return _filter_by_name_or_id(instances=clusters, name_or_id=name_or_id)


def resolve_workspaces(
    client: ServiceFoundryServiceClient,
    name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
) -> List[Workspace]:
    if not ignore_context and not name_or_id:
        workspace = client.session.get_workspace()
        if workspace:
            workspace = Workspace.from_dict(workspace)
            return [workspace]

    if isinstance(cluster_name_or_id, Cluster):
        clusters = [cluster_name_or_id]
        cluster_name_or_id = clusters[0].id
    else:
        clusters = resolve_clusters(
            client=client, name_or_id=cluster_name_or_id, ignore_context=ignore_context
        )

    if not clusters:
        if cluster_name_or_id:
            raise ValueError(f"No cluster found with name or id {cluster_name_or_id!r}")
        else:
            workspaces = [Workspace.from_dict(w) for w in client.list_workspace()]
    elif len(clusters) > 1:
        raise ValueError(
            f"More than one cluster found with name or id {cluster_name_or_id!r}: {clusters!r}"
        )
    else:
        cluster = clusters[0]
        _workspaces = client.get_workspace_by_name(
            workspace_name="", cluster_id=cluster.id
        )
        workspaces = [Workspace.from_dict(w) for w in _workspaces]
    return _filter_by_name_or_id(instances=workspaces, name_or_id=name_or_id)


def resolve_services(
    client: ServiceFoundryServiceClient,
    name_or_id: Optional[str] = None,
    workspace_name_or_id: Optional[Union[Workspace, str]] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
) -> List[Service]:
    if isinstance(workspace_name_or_id, Workspace):
        workspaces = [workspace_name_or_id]
        workspace_name_or_id = workspaces[0].id
    else:
        workspaces = resolve_workspaces(
            client=client,
            name_or_id=workspace_name_or_id,
            cluster_name_or_id=cluster_name_or_id,
            ignore_context=ignore_context,
        )
    if not workspaces:
        if workspace_name_or_id:
            raise ValueError(
                f"No workspaces found with name or id {workspace_name_or_id!r}"
            )
        else:
            services = [Service.from_dict(s) for s in client.list_service()]
    elif len(workspaces) > 1:
        raise ValueError(
            f"More than one workspace found with name or id {workspace_name_or_id!r}: {workspaces!r}"
        )
    else:
        workspace = workspaces[0]
        _services = client.list_service_by_workspace(workspace_id=workspace.id)
        services = [Service.from_dict(s) for s in _services]
    return _filter_by_name_or_id(instances=services, name_or_id=name_or_id)


def resolve_deployments(
    client: ServiceFoundryServiceClient,
    name_or_id: Optional[str] = None,
    service_name_or_id: Optional[Union[Service, str]] = None,
    workspace_name_or_id: Optional[Union[Workspace, str]] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
) -> List[Deployment]:
    if isinstance(service_name_or_id, Service):
        services = [service_name_or_id]
        service_name_or_id = services[0].id
    else:
        services = resolve_services(
            client=client,
            name_or_id=service_name_or_id,
            workspace_name_or_id=workspace_name_or_id,
            cluster_name_or_id=cluster_name_or_id,
            ignore_context=ignore_context,
        )
    if not services:
        if service_name_or_id:
            raise ValueError(
                f"No services found with name or id {service_name_or_id!r}"
            )
        else:
            # TODO (chiragjn): Needs an API from server side to fetch all deployments.
            #                  Otherwise we can do it slowly by doing
            #                  flatten([c.list_deployment(service_id=s.id) for s in all_services])
            raise NotImplementedError(
                "Cannot fetch deployments without a service name or id"
            )
    elif len(services) > 1:
        raise ValueError(
            f"More than one service found with name or id {service_name_or_id!r}: {services!r}"
        )
    else:
        service = services[0]
        _deployments = client.list_deployment(service_id=service.id)
        deployments = [Deployment.from_dict(d) for d in _deployments]
    return _filter_by_name_or_id(instances=deployments, name_or_id=name_or_id)


def ask_pick_cluster(clusters: List[Cluster]) -> Cluster:
    choices = [
        questionary.Choice(title=f"{c.name} ({c.fqn})", value=c) for c in clusters
    ]
    return questionary.select("Pick a cluster", choices=choices).ask()


def maybe_ask_pick_cluster(clusters: List[Cluster]) -> Cluster:
    if len(clusters) == 1:
        return clusters[0]
    return ask_pick_cluster(clusters=clusters)


def ask_pick_workspace(workspaces: List[Workspace]) -> Workspace:
    choices = [
        questionary.Choice(title=f"{w.name} ({w.fqn})", value=w) for w in workspaces
    ]
    return questionary.select("Pick a workspace", choices=choices).ask()


def maybe_ask_pick_workspace(workspaces: List[Workspace]) -> Workspace:
    if len(workspaces) == 1:
        return workspaces[0]
    return ask_pick_workspace(workspaces=workspaces)


def ask_pick_service(services: List[Service]) -> Service:
    choices = [
        questionary.Choice(title=f"{s.name} ({s.fqn})", value=s) for s in services
    ]
    return questionary.select("Pick a service", choices=choices).ask()


def maybe_ask_pick_service(services: List[Service]) -> Service:
    if len(services) == 1:
        return services[0]
    return ask_pick_service(services=services)


def ask_pick_deployment(deployments: List[Deployment]) -> Deployment:
    choices = [
        questionary.Choice(title=f"{d.name} ({d.fqn})", value=d) for d in deployments
    ]
    return questionary.select("Pick a deployment", choices=choices).ask()


def maybe_ask_pick_deployment(deployments: List[Deployment]) -> Deployment:
    if len(deployments) == 1:
        return deployments[0]
    return ask_pick_deployment(deployments=deployments)


def resolve_cluster_or_error(
    name_or_id: Optional[str] = None,
    ignore_context: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Cluster:
    if non_interactive:
        if ignore_context and not name_or_id:
            raise ValueError(
                "cluster name or id cannot be null. Pass it using `--cluster` or `-c`"
            )

    clusters = resolve_clusters(
        client=client, name_or_id=name_or_id, ignore_context=ignore_context
    )

    if not clusters:
        if name_or_id:
            raise ValueError(f"No cluster found with name or id {name_or_id!r}")
        else:
            raise ValueError(f"No clusters found!")
    else:
        if non_interactive:
            if len(clusters) > 1:
                raise ValueError(
                    f"More than one cluster found with name or id {name_or_id!r}: {clusters!r}"
                )
            else:
                cluster = clusters[0]
        else:
            cluster = maybe_ask_pick_cluster(clusters=clusters)
    return cluster


def resolve_workspace_or_error(
    name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Tuple[Workspace, Cluster]:
    if non_interactive:
        if ignore_context and not name_or_id:
            raise ValueError(
                "workspace name or id cannot be null. Pass it using `--workspace` or `-w`"
            )

    if isinstance(cluster_name_or_id, Cluster):
        cluster = cluster_name_or_id
    else:
        cluster = resolve_cluster_or_error(
            name_or_id=cluster_name_or_id,
            non_interactive=non_interactive,
            ignore_context=ignore_context,
            client=client,
        )

    console.print(PROMPT_USING_CLUSTER_CONTEXT.format(cluster.name))

    workspaces = resolve_workspaces(
        client=client,
        name_or_id=name_or_id,
        cluster_name_or_id=cluster,
        ignore_context=ignore_context,
    )
    if not workspaces:
        if name_or_id:
            raise ValueError(
                f"No workspace found with name or id {name_or_id!r} in cluster {cluster.name!r}"
            )
        else:
            raise ValueError(f"No workspaces found!")
    else:
        if non_interactive:
            if len(workspaces) > 1:
                raise ValueError(
                    f"More than one workspace found with name or id {name_or_id!r}: {workspaces!r}"
                )
            else:
                workspace = workspaces[0]
        else:
            workspace = maybe_ask_pick_workspace(workspaces=workspaces)
    return workspace, cluster


def resolve_service_or_error(
    name_or_id: str,
    workspace_name_or_id: Optional[Union[Workspace, str]] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Tuple[Service, Workspace, Cluster]:
    if non_interactive:
        if ignore_context and not name_or_id:
            raise ValueError("service name or id cannot be null")

    if isinstance(cluster_name_or_id, Cluster):
        cluster = cluster_name_or_id
    else:
        cluster = resolve_cluster_or_error(
            name_or_id=cluster_name_or_id,
            non_interactive=non_interactive,
            ignore_context=ignore_context,
            client=client,
        )

    if isinstance(workspace_name_or_id, Workspace):
        workspace = workspace_name_or_id
    else:
        workspace, cluster = resolve_workspace_or_error(
            name_or_id=workspace_name_or_id,
            cluster_name_or_id=cluster,
            non_interactive=non_interactive,
            ignore_context=ignore_context,
            client=client,
        )

    console.print(PROMPT_USING_WORKSPACE_CONTEXT.format(workspace.name))

    services = resolve_services(
        client=client,
        name_or_id=name_or_id,
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        ignore_context=ignore_context,
    )

    if not services:
        if name_or_id:
            raise ValueError(
                f"No service found with name or id {name_or_id!r} in workspace {workspace.name!r}"
            )
        else:
            raise ValueError(f"No services found!")
    else:
        if non_interactive:
            if len(services) > 1:
                raise ValueError(
                    f"More than one service found with name or id {name_or_id!r}: {services!r}"
                )
            else:
                service = services[0]
        else:
            service = maybe_ask_pick_service(services=services)

    return service, workspace, cluster


def resolve_deployment_or_error(
    name_or_id: str,
    service_name_or_id: Optional[Union[Service, str]] = None,
    workspace_name_or_id: Optional[Union[Workspace, str]] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Tuple[Deployment, Service, Workspace, Cluster]:
    if non_interactive:
        if ignore_context and not name_or_id:
            raise ValueError("deployment name or id cannot be null")

    # TODO: resolve separately if we start to have a get all deployments API that can give deployments across services
    service, workspace, cluster = resolve_service_or_error(
        name_or_id=service_name_or_id,
        workspace_name_or_id=workspace_name_or_id,
        cluster_name_or_id=cluster_name_or_id,
        ignore_context=ignore_context,
        non_interactive=non_interactive,
        client=client,
    )

    deployments = resolve_deployments(
        client=client,
        service_name_or_id=service,
        name_or_id=name_or_id,
        workspace_name_or_id=workspace,
        cluster_name_or_id=cluster,
        ignore_context=ignore_context,
    )

    if not deployments:
        if name_or_id:
            raise ValueError(
                f"No deployment found with name or id {name_or_id!r} for service {service.name!r}"
            )
        else:
            raise ValueError(f"No deployments found!")
    else:
        if non_interactive:
            if len(deployments) > 1:
                raise ValueError(
                    f"More than one deployment found with name or id {name_or_id!r}: {deployments!r}"
                )
            else:
                deployment = deployments[0]
        else:
            deployment = maybe_ask_pick_deployment(deployments=deployments)

    return deployment, service, workspace, cluster



import os.path
import sys
import threading
import urllib
import webbrowser
from subprocess import Popen

from ..model.build_pack import BuildPack
from ..output_callback import OutputCallBack
from ..util import execute, run_process


class ClosableProcessWrapperThread(threading.Thread):
    def __init__(self, process: Popen, callback):
        super(ClosableProcessWrapperThread, self).__init__()
        self.process: Popen = process
        self.callback: OutputCallBack = callback

    def stop(self):
        self.process.terminate()

    def run(self):
        pid = self.process.pid
        line = self.process.stdout.readline()
        while line:
            line = self.process.stdout.readline()
            self.callback.print_line(f"[{pid}] {line}")
        self.callback.print_line("Process Stopped.")


def deploy(
    build_pack: BuildPack, component, package_dir, build_dir, callback: OutputCallBack
):
    virtualenv = f"{build_dir}/virtualenv.pyz"
    if not os.path.isfile(virtualenv):
        callback.print_header("Going to download virtualenv")
        urllib.request.urlretrieve(
            "https://bootstrap.pypa.io/virtualenv.pyz", virtualenv
        )

    python_location = sys.executable

    venv = f"{build_dir}/venv"
    if not os.path.isdir(venv):
        callback.print_header("Going to create virtualenv")
        cmd = [python_location, virtualenv, venv]
        for line in execute(cmd):
            callback.print_line(line)

    callback.print_header("Going to install dependency")
    for requirement_file in build_pack.local.requirement_files:
        cmd = [f"{venv}/bin/pip", "install", "-r", f"{package_dir}/{requirement_file}"]
        for line in execute(cmd):
            callback.print_line(line.rstrip())

    callback.print_header(f"Going to run local service in background.\n")
    command = build_pack.local.run_command
    command = f"{venv}/bin/{command}"
    callback.print_line(f"Going to execute command {command}")
    cmd = command.split(" ")
    thread = ClosableProcessWrapperThread(run_process(cmd), callback)
    thread.start()

    url = f"http://127.0.0.1:{component['spec']['container']['ports'][0]['containerPort']}"
    callback.print_line(f"Service will be up on {url}")
    webbrowser.open(url)

    return thread



from typing import List

from ipywidgets import Box, Button, Layout, widgets

from servicefoundry.build.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.build.exceptions import BadRequestException, ConfigurationException
from servicefoundry.build.parser.sf_template import SfTemplate
from servicefoundry.build.parser.template_parameters import (
    NUMBER,
    OPTIONS,
    STRING,
    WORKSPACE,
    TemplateParameter,
)
from servicefoundry.build.session_factory import get_session

style = {"description_width": "350px"}


def init():
    try:
        get_session()
    except BadRequestException:
        return "Please login using login() function before using init."
    tfs_client = ServiceFoundryServiceClient.get_client()
    templates = tfs_client.get_templates_list()
    template_choices = [("Select a template.", None)]
    for t in templates:
        template_choices.append((f'{t["id"]} - {t["description"]}', t["id"]))

    items_layout = Layout(width="auto")
    items = []

    template_dropdown = widgets.Dropdown(
        options=template_choices,
        description="Template:",
        layout=items_layout,
        style=style,
    )
    items.append(template_dropdown)
    box_layout = Layout(
        display="flex", flex_flow="column", align_items="stretch", width="100%"
    )
    box = Box(children=items, layout=box_layout)
    output = widgets.Output()

    def _on_change(change):
        if (
            change["name"] is not None
            and change["name"] == "value"
            and (change["new"] != change["old"])
        ):
            template = change["new"]
            if template is not None:
                template_dropdown.disabled = True
                output.append_stdout(f"You have selected template: {template}\n")
                sf_template = SfTemplate.get_template(f"truefoundry.com/v1/{template}")
                items.extend(
                    _get_form_item(sf_template, items_layout, tfs_client, output)
                )
                box.children = items

    template_dropdown.observe(_on_change)
    return box


# Return form items with click button
def _get_form_item(sf_template: SfTemplate, layout, tfs_client, output):
    items, fetchers = _get_items(sf_template.parameters, layout, tfs_client)
    button = Button(description="Initialize", tooltip="Initialize", layout=layout)

    def on_button_clicked(_):
        try:
            parameter_values = {}
            for fetcher in fetchers:
                key, value = fetcher()
                parameter_values[key] = value
            project_folder = sf_template.generate_project(parameter_values)
            output.append_stdout(
                f"New project is initialized in directory {project_folder}.\n"
            )
            output.append_stdout(
                f"Use sfn.deploy(project_folder='{project_folder}', local=True) to run locally.\n"
            )
        except ConfigurationException as err:
            output.append_stderr(f"{err.message}\n")
            return
        button.disabled = True

    button.on_click(on_button_clicked)
    items.append(button)
    items.append(output)
    return items


def _input_fetcher(param, item):
    def fetcher():
        value = item.value
        if param.kind == NUMBER:
            if value.isnumeric():
                value = int(value)
            else:
                msg = f"Answer of {param.prompt} should be integer."
                raise ConfigurationException(msg)
        return param.id, value

    return fetcher


def _dropdown_fetcher(param, item):
    def fetcher():
        value = item.value
        return param.id, value

    return fetcher


def _get_items(parameters: List[TemplateParameter], layout, tfs_client):
    items = []
    fetchers = []
    for param in parameters:
        kind = param.kind
        if kind in [STRING, NUMBER]:
            item = widgets.Text(
                value=str(param.default),
                placeholder=str(param.default),
                description=param.prompt,
                disabled=False,
                layout=layout,
                style=style,
            )
            fetcher = _input_fetcher(param, item)

        elif kind == WORKSPACE:
            item = widgets.Dropdown(
                options=param.get_workspaces(tfs_client),
                description=param.prompt,
                layout=layout,
                style=style,
            )
            fetcher = _dropdown_fetcher(param, item)
        elif kind == OPTIONS:
            item = widgets.Dropdown(
                options=param.options,
                description=param.prompt,
                layout=layout,
                style=style,
            )
            fetcher = _dropdown_fetcher(param, item)
        else:
            raise RuntimeError(f"Unhandled template parameter kind {kind}")
        items.append(item)
        fetchers.append(fetcher)
    return items, fetchers
